/* Team members: XXX
 */

#include <assert.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include <signal.h>
#include <sys/errno.h>
#include <sys/wait.h>
#include <unistd.h>

/* given a newline-terminated string of input, produce an array of the space-separated words from that input

   the input string is modified in place, replacing each space character with a null (as well as the final newline)

   the returned array will be null terminated, i.e., suitable for execve

   sets num_words to number of words (array is actually one longer, for NULL termination)
*/
char **parse_line(char *line, ssize_t len, int *num_words)
{
    assert(line);
    assert(len > 0);
    assert(line[len - 1] == '\n');

    // count words (no escaping or anything)
    *num_words = 1;
    for (int i = 0; i < len; i += 1)
    {
        if (line[i] == ' ')
        {
            *num_words += 1;
        }
    }

    // allocate entries for each word, plus an extra one for NULL termination
    char **words = (char **)malloc(sizeof(char *) * (*num_words + 1));

    char *current_word = line;
    int word_index = 0;
    for (int i = 0; i < len; i += 1)
    {

        if (line[i] == ' ' || line[i] == '\n')
        {
            // save the current word, null terminated
            line[i] = '\0';
            words[word_index] = current_word;

            word_index += 1;
            current_word = line + i + 1;
        }
    }

    words[*num_words] = NULL;

    return words;
}

typedef struct job
{
    char *command;
    pid_t pid;
    int status;
    struct job *next;
} job_t;

job_t *jobs = NULL;

void add_job(char *command, pid_t pid)
{
    job_t *j = (job_t *)malloc(sizeof(job_t));

    j->command = command;
    j->pid = pid;
    j->status = 0;
    j->next = jobs;

    jobs = j;
}

void free_job(job_t *j)
{
    free(j->command);
    free(j);
}

/* check on jobs

    options will be used to waitpid to, e.g., determine blocking with WNOHANG
*/
void check_jobs(int options)
{
    // TODO #4: check each job in turn, outputing completion/status
}

/* sets up signal handlers */
void init()
{
    struct sigaction action;

    // ignore Ctrl-C
    action.sa_flags = 0;
    action.sa_handler = SIG_IGN;
    sigaction(SIGINT, &action, NULL);

    // TODO #6: handle SIGCHLD
}

int main(int argc, char *argv[])
{
    char *buf = NULL;
    size_t linecap = 0;

    init();

    while (1)
    {
        // TODO #5: check on jobs, update the user
        // TODO #6: only update the user when something has changed

        fprintf(stderr, "¢ ");
        errno = 0;
        ssize_t len = getline(&buf, &linecap, stdin);

        if (len == -1)
        {
            if (!errno)
            {
                // TODO #4: wait for all jobs to finish

                fprintf(stderr, "\nGoodbye!\n");
                return 0;
            }

            perror("I/O error reading line");
            return 1;
        }

        if (len > 1)
        {
            int bg = 0;
            char *command = NULL;

            // see if we're a background task; munge line
            if (buf[len - 2] == '&')
            {
                int last_non_space = len - 3;
                for (; last_non_space > 0; last_non_space -= 1)
                {
                    if (buf[last_non_space] != ' ')
                    {
                        break;
                    }
                }
                buf[last_non_space + 1] = '\n';
                buf[last_non_space + 2] = '\0';
                len = last_non_space + 2; // exclude null!

                bg = 1;

                // save command name, but drop newline
                command = (char *)malloc(sizeof(char) * len);
                strncpy(command, buf, len - 1);
                command[len - 1] = '\0';
            }

            int num_words = 0;
            char **args = parse_line(buf, len, &num_words);
            assert(args);
            assert(args[num_words] == NULL);

            // TODO #1: run the command
            // TODO #2: print the status if it's non-zero
            // TODO #3: run background jobs in the background
        }
    }

    return 0;
}
